#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdint.h>
#include <fcntl.h>

#include "common.h"
#include "macros.h"
#include "io.h"
#include "signal-event.h"
#include "misc.h"
#include "timeout.h"
#include "buffer.h"
#include "hal-proto.h"
#include "hal-peer.h"
#include "hal-cmd.h"

#define info(fmt, arg...) fprintf(stderr, "info: %s: "fmt CRLF, \
				__FUNCTION__, ##arg);

#define error(eval, fmt, arg...)			\
do {							\
	fprintf(stderr, "error: %s: "fmt CRLF,		\
				__FUNCTION__, ##arg);	\
	exit(eval);					\
} while (0)

static struct io_context *io;

static void request_commands(struct hal_peer_base *p)
{
	hal_cmd_connect_request(p->wb);
	hal_cmd_cycle_begin_request(p->wb);
	hal_cmd_cycle_run_request(p->wb);
	hal_cmd_cycle_end_request(p->wb);
	hal_cmd_read_prepare_request(p->wb);
	hal_cmd_read_end_request(p->wb);
	hal_cmd_read_inputs_request(p->wb, 12, 24);
	hal_cmd_read_outputs_request(p->wb, 12, 24);
	hal_cmd_write_prepare_request(p->wb);
	hal_cmd_write_end_request(p->wb);
	hal_cmd_write_inputs_request(p->wb, 24, "00112233");
	hal_cmd_write_outputs_request(p->wb, 24, "00112233");
	hal_cmd_disconnect_request(p->wb);
}

static void reply_commands(struct hal_peer_base *p)
{
	hal_cmd_connect_reply(p->wb, 100, 128, 1);
	hal_cmd_cycle_begin_reply(p->wb);
	hal_cmd_cycle_run_reply(p->wb);
	hal_cmd_cycle_end_reply(p->wb);
	hal_cmd_read_prepare_reply(p->wb);
	hal_cmd_read_end_reply(p->wb);
	hal_cmd_read_inputs_reply(p->wb, 24, "44556677");
	hal_cmd_read_outputs_reply(p->wb, 24, "44556677");
	hal_cmd_write_prepare_reply(p->wb);
	hal_cmd_write_end_reply(p->wb);
	hal_cmd_write_inputs_reply(p->wb);
	hal_cmd_write_outputs_reply(p->wb);
	hal_cmd_disconnect_reply(p->wb);
}

static void peer_event_handler(struct hal_peer_base *peer,
			       hal_peer_event_t event)
{
	switch (event) {
	case HAL_PEER_EVENT_ERROR:
	case HAL_PEER_EVENT_TIMEOUT:
		info("got error or timeout event");
		io_loop_stop(peer->io);
		break;
	case HAL_PEER_EVENT_WRITE:
		if (buffer_size(peer->wb) == 0)
			io_loop_stop(peer->io);
		break;
	case HAL_PEER_EVENT_READ:
		break;
	}
}

static void handle_signals(int signo, siginfo_t *si, void *opaque)
{
	switch (signo) {
	case SIGINT:
	case SIGTERM:
		info("exiting by signal %d", signo);
		io_loop_stop(io);
		break;
	}
}

int main(int argc, char **argv)
{
	struct hal_peer_base *p;
	int rfd, wfd;
	ret_t ret;

	if (argc < 2)
		error(1, "not enough arguments: write fifo, read fifo");

	/* Open pipes and get file descriptors. */
	ret = pipe_open(argv[1], O_RDWR, &wfd);
	if (ret != ret_ok)
		error(1, "can't open write pipe");

	ret = pipe_open(argv[2], O_RDWR, &rfd);
	if (ret != ret_ok)
		error(1, "can't open read pipe");

	/* Set signal handlers. */
	sigevent_ignore(SIGPIPE);
	sigevent_add(SIGINT, handle_signals, NULL, 1);
	sigevent_add(SIGTERM, handle_signals, NULL, 1);
	sigevent_init();

	/* Init loop facilities. */
	timeout_init();
	io = io_new();

	/* Create a single peer. */
	p = hal_peer_base_new();
	hal_peer_base_init(p, io, wfd, rfd, peer_event_handler);

	/* First send all request commands then send all reply commands. */
	request_commands(p);
	reply_commands(p);

	hal_peer_send_default(p);
	/* Parse commands. */
	io_loop_run(io);

	/* Free resources. */
	hal_peer_base_deinit(p);
	hal_peer_base_free(&p);

	io_free(&io);
	timeout_deinit();
	sigevent_deinit();

	return 0;
}

